/*
 * Copyright (C) 2019 - 2020  Magnus Bergman <magnus@snisurset.net>
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, see <http:/www.gnu.org/licenses/>
 */
#include <abydos-plugin.h>
#include <assert.h>
#include <debug.h>
#include <stdint.h>                     /* needed by libbpg.h */
#include <libbpg.h>
#include <stdlib.h>                     /* malloc(), free() */


typedef struct{
    cairo_surface_t *surface;
    double duration;
}frame_t;

struct _abydos_plugin_handle_t{
    abydos_plugin_info_t *info;
    frame_t *frame;
};


static const char **
_bpg_get_mime_types()
{
    static const char *mime_types[] = {
        "image/bpg",
        "image/x-bpg",
        NULL
    };
    return mime_types;
}


static abydos_plugin_handle_t *
_bpg_new(const char *mime_type, abydos_plugin_info_t *info)
{
    abydos_plugin_handle_t *h = (abydos_plugin_handle_t*)malloc(sizeof(abydos_plugin_handle_t));
    h->info = info;
    h->info->threadsafe = 1;
    h->frame = NULL;
    return h;
}


static void
_bpg_free(abydos_plugin_handle_t *h)
{
    if(h->frame){
        int i;
        for(i=0;i<h->info->frame_count;++i)
            cairo_surface_destroy(h->frame[i].surface);
        free(h->frame);
    }
    free(h);
}


static cairo_surface_t *
_surface_from_context(BPGDecoderContext *ctx, int width, int height, int alpha, int premultiply)
{
    cairo_surface_t *surface = cairo_image_surface_create(alpha ? CAIRO_FORMAT_ARGB32 : CAIRO_FORMAT_RGB24,width,height);
    uint8_t *line = (uint8_t*)malloc(width*4);
    uint32_t *dst = (uint32_t*)cairo_image_surface_get_data(surface);
    int dst_skip = cairo_image_surface_get_stride(surface)/sizeof(uint32_t)-width;

    if(premultiply){
        while(height-->0){
            const uint8_t *src = line;
            int x;
            bpg_decoder_get_line(ctx, line);
            for(x=width;x>0;--x){
                int r = *src++;
                int g = *src++;
                int b = *src++;
                int a = *src++;
                if(a<255){
                    r = r*a/255;
                    g = g*a/255;
                    b = b*a/255;
                }
                *dst++ = r<<16|g<<8|b<<0|a<<24;
            }
            dst += dst_skip;
        }
    }else{
        while(height-->0){
            const uint8_t *src = line;
            int x;
            bpg_decoder_get_line(ctx, line);
            for(x=width;x>0;--x){
                int r = *src++;
                int g = *src++;
                int b = *src++;
                int a = *src++;
                *dst++ = r<<16|g<<8|b<<0|a<<24;
            }
            dst += dst_skip;
        }
    }

    free(line);
    cairo_surface_mark_dirty(surface);
    return surface;
}


static int
_bpg_create_from_data(abydos_plugin_handle_t *h,const char *bytes, size_t len)
{
    int result = -1;
    BPGDecoderContext *ctx = bpg_decoder_open();
    BPGImageInfo info;
    if(bpg_decoder_decode(ctx,(const uint8_t*)bytes,len)<0)
        goto out;
    if(bpg_decoder_get_info(ctx,&info)<0)
        goto out;
    h->info->width = info.width;
    h->info->height = info.height;
    h->info->frame_count = 0;
    for(;;){
        int delay_num;
        int delay_den;
        if(bpg_decoder_start(ctx, BPG_OUTPUT_FORMAT_RGBA32) < 0)
            break;
        bpg_decoder_get_frame_duration(ctx, &delay_num, &delay_den);
        h->frame = realloc(h->frame, sizeof(frame_t) * (h->info->frame_count+1));
        h->frame[h->info->frame_count].surface
            = _surface_from_context(ctx, h->info->width, h->info->height, info.has_alpha, info.has_alpha && !info.premultiplied_alpha);
        h->frame[h->info->frame_count].duration = (double)delay_num/delay_den;
        ++h->info->frame_count;
    }

    result = 0;
out:
    bpg_decoder_close(ctx);
    return result;
}


static cairo_surface_t *
_bpg_get_image_surface(abydos_plugin_handle_t *h,
                       int page,
                       cairo_bool_t *layer_visibility,
                       int variant,
                       double time,
                       int frame)
{
    return cairo_surface_reference(h->frame[frame].surface);
}


static void
_bpg_get_duration(abydos_plugin_handle_t *h,int page,double *duration)
{
    int i;
    DEBUG_p("begin");
    for(i=0;i<h->info->frame_count;++i)
        duration[i] = h->frame[i].duration;
}


const abydos_plugin_interface_t *
Abydos0()
{
    static const abydos_plugin_interface_t p = {
        .version = ABYDOS_PLUGIN_INTERFACE_VERSION,
        .get_mime_types = _bpg_get_mime_types,
        .new = _bpg_new,
        .free = _bpg_free,
        .create_from_data = _bpg_create_from_data,
        .get_image_surface = _bpg_get_image_surface,
        .get_duration = _bpg_get_duration
    };
    return &p;
}
