/*
 * Copyright (C) 2020  Magnus Bergman <magnus@snisurset.net>
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, see <http:/www.gnu.org/licenses/>
 */
#define PY_SSIZE_T_CLEAN
#include <abydos-plugin.h>
#include <Imaging.h>
#include <Python.h>
#include <stdlib.h>
#include <string.h>

#include "pixels.h"

/*

Type code   | write |
------------+-------<
  BLP       | no    |
  BMP       | yes   |
  CUR       | no    |
  DCX       | no    |
  DDS       | no    |
  DIB       | yes   |
* EPS       | yes   |
  FLI, FLC  | no    |
* FPX       | no    |
  FTEX      | no    |
  GBR       | no    |
  GD        | no    |
* GIF       | yes   |
  ICNS      | yes   |
  ICO       | yes   |
* IM        | yes   |
  IMT       | no    |
  IPTC/NAA  | no    |
* JPEG      | yes   |
* JPEG 2000 | yes   |
  MCIDAS    | no    |
  MIC       | no    |
  MPO       | no    |
  MSP       | yes   |
  PCD       | no    |
* PCX       | yes   |
  PIXAR     | no    |
* PNG       | yes   |
  PPM       | yes   |
  PSD       | no    |
  SGI       | yes   |
* SPIDER    | yes   |
* TGA       | yes   |
* TIFF      | yes   |
  WAL       | no    |
  WebP      | yes   |
  WMF       | no    |
  XBM       | yes   |
  XPM       | no    |
------------^-------'

*/

typedef struct{
    cairo_surface_t *surface;
    double duration;
}frame_t;

struct _abydos_plugin_handle_t{
    abydos_plugin_info_t *info;
    frame_t *frame;
};

typedef struct{
    PyObject_HEAD
    Imaging image;
}image_object_t;


static const char **
_pil_get_mime_types()
{
    static const char *mime_types[] = {
        "image/x-eps",
        "image/vnd.fpx",
        "image/gif",
        "image/jpeg",
        "image/jp2", "image/jpeg2000", "image/jpx",
        "image/x-labeye-image",
        "image/vnd.zbrush.pcx",
        "image/png",
        "image/x-spider",
        "image/x-tga",
        "image/tiff",
        NULL
    };
    return mime_types;
}


static abydos_plugin_handle_t *
_pil_new(const char *mime_type, abydos_plugin_info_t *info)
{
    abydos_plugin_handle_t *h = (abydos_plugin_handle_t*)malloc(sizeof(abydos_plugin_handle_t));
    h->info = info;
    h->info->threadsafe_render = 1;
    h->frame = NULL;
    return h;
}


static void
_pil_free(abydos_plugin_handle_t *h)
{
    free(h);
}


static void
_convert_palette(uint32_t *dst, ImagingPalette img_palette)
{
    abydos_pixel_format_t format;
    const uint8_t *src;
    int i;
    if(!strcmp(img_palette->mode, "RGBA")){
        format = ABYDOS_PIXEL_FORMAT_r8g8b8a8;
    }else{
        return;
    }
    src = img_palette->palette;
    for(i=0;i<256;++i){
        int a = src[((channel_order_t)format).a];
        int r = a*src[((channel_order_t)format).r]/255;
        int g = a*src[((channel_order_t)format).g]/255;
        int b = a*src[((channel_order_t)format).b]/255;
        *dst++ = r<<16|g<<8|b<<0|a<<24;
        src += 4;
    }
}


static cairo_surface_t *
_surface_from_image(Imaging img)
{
    abydos_pixel_format_t format;
    int channel_width;
    uint32_t palette[256];
    cairo_surface_t *surface;
    uint32_t *dst;
    int dst_skip;
    int x;
    int y;
    if(!strcmp(img->mode, "F")){
        format = ABYDOS_PIXEL_FORMAT_g32;
    }else if(!strcmp(img->mode, "P")){
        format = ABYDOS_PIXEL_FORMAT_c8a;
        _convert_palette(palette, img->palette);
    }else if(!strcmp(img->mode, "RGB")){
        format = ABYDOS_PIXEL_FORMAT_r8g8b8;
    }else if(!strcmp(img->mode, "RGBA")){
        format = ABYDOS_PIXEL_FORMAT_r8g8b8a8;
    }else{
        return NULL;
    }
    channel_width = img->pixelsize / img->bands;
    surface = cairo_image_surface_create(has_alpha(format) ? CAIRO_FORMAT_ARGB32 : CAIRO_FORMAT_RGB24,img->xsize,img->ysize);
    dst = (uint32_t*)cairo_image_surface_get_data(surface);
    dst_skip = cairo_image_surface_get_stride(surface)/sizeof(uint32_t)-img->xsize;
    switch(channel_width){
    case 1:
        if(format<0){
            for(y=0;y<img->ysize;++y){
                const uint8_t *src = (uint8_t*)img->image[y];
                for(x=0;x<img->xsize;++x)
                    *dst++ = palette[*src++];
                dst += dst_skip;
            }
        }else if(has_alpha(format)){
            for(y=0;y<img->ysize;++y){
                const uint8_t *src = (uint8_t*)img->image[y];
                for(x=0;x<img->xsize;++x){
                    int a = src[((channel_order_t)format).a];
                    int r = a*src[((channel_order_t)format).r]/255;
                    int g = a*src[((channel_order_t)format).g]/255;
                    int b = a*src[((channel_order_t)format).b]/255;
                    *dst++ = r<<16|g<<8|b<<0|a<<24;
                    src += img->pixelsize;
                }
                dst += dst_skip;
            }
        }else{
            for(y=0;y<img->ysize;++y){
                const uint8_t *src = (uint8_t*)img->image[y];
                for(x=0;x<img->xsize;++x){
                    int r = src[((channel_order_t)format).r];
                    int g = src[((channel_order_t)format).g];
                    int b = src[((channel_order_t)format).b];
                    *dst++ = r<<16|g<<8|b<<0;
                    src += img->pixelsize;
                }
                dst += dst_skip;
            }
        }
        break;
    case 4:
        if(is_gray(format) && has_alpha(format)){
        }else if(is_gray(format)){
            for(y=0;y<img->ysize;++y){
                const uint8_t *src = (uint8_t*)img->image[y];
                for(x=0;x<img->xsize;++x){
                    int g = *(float*)&src[((channel_order_t)format).g];
                    *dst++ = g<<16|g<<8|g<<0;
                    src += img->pixelsize;
                }
                dst += dst_skip;
            }
        }else if(has_alpha(format)){
            for(y=0;y<img->ysize;++y){
                const uint8_t *src = (uint8_t*)img->image[y];
                for(x=0;x<img->xsize;++x){
                    int r = *(float*)&src[((channel_order_t)format).r];
                    int g = *(float*)&src[((channel_order_t)format).g];
                    int b = *(float*)&src[((channel_order_t)format).b];
                    *dst++ = r<<16|g<<8|b<<0;
                    src += img->pixelsize;
                }
                dst += dst_skip;
            }
        }else{
            for(y=0;y<img->ysize;++y){
                const uint8_t *src = (uint8_t*)img->image[y];
                for(x=0;x<img->xsize;++x){
                    float a = *(float*)&src[((channel_order_t)format).a];
                    int r = a * *(float*)&src[((channel_order_t)format).r] / 255;
                    int g = a * *(float*)&src[((channel_order_t)format).g] / 255;
                    int b = a * *(float*)&src[((channel_order_t)format).b] / 255;
                    *dst++ = r<<16|g<<8|b<<0|((int)a)<<24;
                    src += img->pixelsize;
                }
                dst += dst_skip;
            }
        }
        break;
    }
    cairo_surface_mark_dirty(surface);
    return surface;
}


static int
_pil_create_from_data(abydos_plugin_handle_t *h,const char *data, size_t len)
{
    PyObject *mod;
    PyObject *bytes;
    PyObject *io;
    PyObject *img;
    PyObject *pixels;
    PyObject *im;
    Imaging image;

    Py_Initialize();

    bytes = PyBytes_FromStringAndSize(data, len);

    mod = PyImport_ImportModule("io");
    io = PyObject_CallMethod(mod, "BytesIO", "O", bytes);

    mod = PyImport_ImportModule("PIL.Image");
    img = PyObject_CallMethod(mod, "open", "O", io);
    pixels = PyObject_CallMethod(img, "load", "");
    Py_DECREF(pixels);
    im = PyObject_GetAttrString(img, "im");
    image = ((image_object_t*)im)->image;

    h->info->width = image->xsize;
    h->info->height = image->ysize;
    printf("width: %d, height: %d\n", h->info->width, h->info->height);

    h->frame = (frame_t*)malloc(h->info->frame_count * sizeof(frame_t));
    h->frame[0].surface = _surface_from_image(image);

    Py_DECREF(im);
    Py_DECREF(img);
    Py_DECREF(io);
    Py_DECREF(bytes);
    Py_Finalize();
    return 0;
}


static cairo_surface_t *
_pil_get_image_surface(abydos_plugin_handle_t *h,
                       int page,
                       cairo_bool_t *layer_visibility,
                       int variant,
                       double time,
                       int frame)
{
    return cairo_surface_reference(h->frame[frame].surface);
}


static void
_pil_get_duration(abydos_plugin_handle_t *h, int page, double *duration)
{
}


const abydos_plugin_interface_t *
Abydos0()
{
    static const abydos_plugin_interface_t p = {
        .version = ABYDOS_PLUGIN_INTERFACE_VERSION,
        .get_mime_types = _pil_get_mime_types,
        .new = _pil_new,
        .free = _pil_free,
        .create_from_data = _pil_create_from_data,
        .get_image_surface = _pil_get_image_surface,
        .get_duration = _pil_get_duration
    };
    return &p;
}
