/*
 * Copyright (C) 2020  Magnus Bergman <magnus@snisurset.net>
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, see <http:/www.gnu.org/licenses/>
 */
#include <cstring>

#include "xpk.h"

extern "C"{
#define new new_
#include <abydos-plugin.h>
#undef new
#include <debug.h>

#include "amiga.h"
#include "pixels.h"
}

/*

Header:

-------v------v--------------
offset | size | Description
-------+------+--------------
     0 |   18 | "SVG Graphics File\0"
    18 |    2 | Version (always 1)
    20 |    4 | Header size
    24 |    4 | X offset
    28 |    4 | Y offset
    32 |    4 | Width
    36 |    4 | Height
    40 |    4 | Depth
    44 |    4 | View mode
    48 |    1 | Bits
    49 |    1 | Planes
    50 |    4 | Stride
    54 |  768 | Palette

*/

#define ID_XPKF ID('X','P','K','F')
#define ID_PP20 ID('P','P','2','0')
#define ID_LZ77 ID('L','Z','7','7')

struct _abydos_plugin_handle_t{
    abydos_plugin_info_t *info;
    cairo_surface_t *surface;
};


#define get8(a) (*(uint8_t*)(a))
#define get16(a) be16toh(*(uint16_t*)(a))
#define get32(a) be32toh(*(uint32_t*)(a))


static void
_expand_palette(const uint8_t *src, uint32_t *dst, int colors)
{
    while(colors-->0){
        int r = *src++;
        int b = *src++;
        int g = *src++;
        *dst++ = r<<16|g<<8|b<<0|255<<24;
    }
}


static int
_unpack_xpk_chunk(xpk_compression_t comp, const uint8_t *src, size_t src_len, const uint8_t *dst_start, uint8_t *dst, size_t dst_len)
{
    ConstStaticBuffer src_buf(src, src_len);
    MemoryBuffer dst_buf(dst, dst_len);
    ConstStaticBuffer prev_buf(dst_start, dst-dst_start);
    std::unique_ptr<XPKDecompressor::State> state;
    XPKDecompressor *dec = xpk_lookup_decompressor(comp, src_buf, state);
    if(!dec)
        return -1;
    int ret = 0;
    try{
        dec->decompressImpl(dst_buf, prev_buf, true);
    }catch(const std::exception&){
        DEBUG_p("chunk decompression failed");
        ret = -1;
    }
    delete dec;
    return ret;
}


static uint8_t *
_unpack_xpk(const uint8_t *src, const uint8_t *src_end, int dst_len)
{
    DEBUG_p("XPKF compression (%d bytes)",(int)(src_end-src));

    if(src+40 > src_end)
        return NULL;

    size_t packed_size = get32(src+0);
    if(src+4+packed_size > src_end)
        return NULL;

    size_t unpacked_size = get32(src+8);
    if(dst_len > unpacked_size)
        return NULL;

    int flags = get8(src+28);
    if(flags&2){
        DEBUG_p("password protected");
        return NULL;
    }
    int header_len = flags&1 ? 12 : 8;

    xpk_compression_t comp = (xpk_compression_t)*(uint32_t*)(src+4);

    if(flags&4)
        src += 2+get16(src+32);
    src += 32;

    uint8_t *body;
    uint8_t *dst = body = new uint8_t[unpacked_size];
    const uint8_t *dst_end = dst+unpacked_size;

    for(;;){
        int chunk_type = get8(src);
        int packed_chunk_size;
        int unpacked_chunk_size;
        if(flags&1){
            packed_chunk_size = get32(src+4);
            unpacked_chunk_size = get32(src+8);
        }else{
            packed_chunk_size = get16(src+4);
            unpacked_chunk_size = get16(src+6);
        }

        if(src+header_len+packed_chunk_size > src_end)
            goto error;
        if(dst+unpacked_chunk_size > dst_end)
            goto error;

        src += header_len;
        if(0==chunk_type){
            /* uncompressed data */
            if(packed_chunk_size != unpacked_chunk_size){
                DEBUG_p("packed chunk size (%d) differs from unpacked chunk size (%d)",(int)packed_chunk_size,(int)unpacked_chunk_size);
                goto error;
            }
            memcpy(dst, src, unpacked_chunk_size);
        }else if(1==chunk_type){
            /* compressed data */
            if(_unpack_xpk_chunk(comp, src, packed_chunk_size, body, dst, unpacked_chunk_size)<0)
                goto error;
        }else if(15==chunk_type){
            /* explicit end */
            break;
        }else{
            /* invalid chunk type */
            DEBUG_p("invalid chunk type");
            goto error;
        }

        src += (packed_chunk_size+3)&~3;
        if(src >= src_end)
            break;

        DEBUG_p("bytes left: %d",(int)(src_end-src));

        dst += unpacked_chunk_size;
        if(dst >= dst_end)
            break;
    }
    return body;
error:
    DEBUG_p("unpack error");
    delete body;
    return NULL;
}


static uint8_t *
_copy_uncompressed(const uint8_t *src, int src_len, int dst_len)
{
    uint8_t *body = new uint8_t[dst_len];
    if(dst_len > src_len){
        std::memset(body+src_len,0,dst_len-src_len);
        dst_len = src_len;
    }
    std::memcpy(body, src, dst_len);
    return body;
}


static uint8_t *
_unpack(const uint8_t *src, int src_len, int dst_len)
{
    const uint8_t *end = src+src_len;
    DEBUG_p("packed body is %d bytes",src_len);
    if(src+8 > end){
        DEBUG_p("file too short");
        return NULL;
    }
    switch(*(uint32_t*)src){
    case ID_XPKF:
        DEBUG_p("PKF compression");
        src += 4;
        return _unpack_xpk(src, end, dst_len);
    case ID_PP20:
        DEBUG_p("PP20 compression");
        return NULL;
    case ID_LZ77:
        DEBUG_p("LZ77 compression");
#if 0
        src += 4;
        return _unpack_lz77(src, end, dst_len);
#endif
    default:
        DEBUG_p("unknown compression, assume uncompressed");
        return _copy_uncompressed(src, src_len, dst_len);
    }
}


static int
_sview_create_from_data(abydos_plugin_handle_t *h,const char *data, size_t len)
{
    amiga_cairo_color_t palette[256];
    int bits;
    int planes;
    int depth;
    int stride;
    uint32_t bplcon0;
    uint8_t *body;
    int body_offs;
    int body_len;

    if(len<58)
        return -1;
    if(std::memcmp(data, "SVG Graphics File\0", 18) && memcmp(data, "SGX Graphics File\0", 18))
        return -1;
    if(get16(data+18)!=1)
        return -1;

    bits = get8(data+48);
    planes = get8(data+49);
    if(bits > 1 && planes > 1)
        return -1;

    depth = get32(data+40);
    DEBUG_p("bits: %d, planes: %d, depth: %d",bits,planes,depth);

    switch(bits){
    case 1:
        if(planes > 8 || depth > planes)
            return -1;
        break;
    case 8:
        if(depth > 8)
            return -1;
        break;
    case 24:
    case 32:
    case 48:
    case 64:
        if(bits != depth)
            return -1;
        break;
    default:
        return -1;
    }

    if(depth <= 8){
        if(54 + (1<<depth) > len)
            return -1;
        _expand_palette((uint8_t*)data+54,palette,1<<depth);
    }

    h->info->width = get32(data+32);
    h->info->height = get32(data+36);
    stride = get32(data+50);
    DEBUG_p("size: %dx%d, stride: %d",h->info->width,h->info->height, stride);

    body_offs = get32(data+20);
    DEBUG_p("data offset: %d",body_offs);
    body_len = stride * h->info->height * planes;

    bplcon0 = get32(data+44);

    body = _unpack((uint8_t*)data+body_offs, len-body_offs, body_len);
    if(!body)
        return -1;

    DEBUG_p("success");

    /* create surface from body */
    if(1==bits){
        /* bitplanes */
        const uint8_t *bitplane[8];
        int i;
        for(i=0;i<planes;++i)
            bitplane[i] = &body[stride*h->info->height*i];
        h->surface = abydos_image_surface_create_from_amiga_bitplanes(bitplane, h->info->width, h->info->height, depth, bplcon0, palette, 1<<depth, 0);
    }else if(8==bits){
        h->surface = abydos_image_surface_create_from_pixels(ABYDOS_PIXEL_FORMAT_c8, 0, body, h->info->width, h->info->height, stride, 1, palette);
    }else{
        abydos_pixel_format_t format;
        int channels;
        switch(depth){
        case 24:
            format = ABYDOS_PIXEL_FORMAT_r8g8b8;
            channels = 3;
            break;
        case 32:
            format = ABYDOS_PIXEL_FORMAT_r8g8b8a8;
            channels = 4;
            break;
        case 48:
            format = ABYDOS_PIXEL_FORMAT_r16g16b16;
            channels = 3;
            break;
        case 64:
            format = ABYDOS_PIXEL_FORMAT_r16g16b16a16;
            channels = 4;
            break;
        }
        h->surface = abydos_image_surface_create_from_pixels(format, 0, body, h->info->width, h->info->height, stride, channels, NULL);
    }
    delete body;
    return 0;
}


extern "C" const abydos_plugin_interface_t *
Abydos0()
{
    static const abydos_plugin_interface_t p = {
        .version = ABYDOS_PLUGIN_INTERFACE_VERSION,
        .get_mime_types = [](){
            static const char *mime_types[] = {
                "image/x-superview",
                NULL
            };
            return mime_types;
        },
        .new_ = [](const char *mime_type, abydos_plugin_info_t *info){
            auto *h = new abydos_plugin_handle_t;
            h->info = info;
            h->info->threadsafe = 1;
            h->surface = NULL;
            return h;
        },
        .free = [](abydos_plugin_handle_t *h){
            if(h->surface)
                cairo_surface_destroy(h->surface);
            delete h;
        },
        .create_from_data = _sview_create_from_data,
        .get_image_surface = [](abydos_plugin_handle_t *h,
                                int page,
                                cairo_bool_t *layer_visibility,
                                int variant,
                                double time,
                                int frame)
        {
            return cairo_surface_reference(h->surface);
        }
    };
    return &p;
}
